/*
 * Copyright (c) 1998 The University of Utah. All rights reserved.
 * Copyright (c) 1996, 1997
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Contributed by the Flux Research Group at the University of Utah.
 * Authors: Godmar Back, Leigh Stoller
 */

/*
 * This file links the ThreadInterface as defined by Tim to the revised
 * version of the internal thread interface ("jthreads").
 *
 * This file implements the functions in ThreadInterface by invoking
 * functions declared in jthread.h
 * It is responsible for unhanding, as well as invoking functions in other
 * parts of the VM.
 *
 * Note that jthreads - while heavily based on Tim's code - can be used 
 * independently of the Kaffe JVM.  That is, they don't have any compile or
 * link time dependencies on it.  This file knows about all the JVM related 
 * types, whereas jthread does not.
 */

#include "debug.h"
#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "config-signal.h"
#include "jtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "lookup.h"
#include "thread.h"
#include "locks.h"
#include "exception.h"
#include "support.h"
#include "external.h"
#include "errors.h"
#include "gc.h"
#include "md.h"
#include "lerrno.h"
#define NOUNIXPROTOTYPES
#include "jsyscall.h"

#include "jthread.h"

/* 
 * store the native thread for the main thread here 
 * between init and createFirst 
 */
static struct Hkaffe_util_Ptr*	mainthread;
static int threadStackSize;	/* native stack size */
/* referenced by native/Runtime.c */
jbool runFinalizerOnExit;	/* should we run finalizers? */

static void 
broadcastDeath(void *jlThread)
{
        Hjava_lang_Thread *tid = jlThread;

        /* Notify on the object just in case anyone is waiting */
        lockMutex(&tid->base);
        broadcastCond(&tid->base);
        unlockMutex(&tid->base);
}

static void 
throwDeath(void)
{
	throwException(ThreadDeath);
}

static void *
thread_malloc(size_t s)
{
	return gc_malloc(s, GC_ALLOC_THREADCTX);
}

static void
thread_free(void *p)
{
	gc_free(p);
}

static
void
runfinalizer(void)
{
	if (runFinalizerOnExit)
		invokeFinalizer();
}

static
void
Tinit(int nativestacksize)
{
	static int inited = 0;
	if (inited)
		return;
	inited++;

	threadStackSize = nativestacksize;
	mainthread = (struct Hkaffe_util_Ptr*)jthread_init(
		DBGEXPR(JTHREADNOPREEMPT, false, true),
		java_lang_Thread_MAX_PRIORITY+1,
		java_lang_Thread_MIN_PRIORITY,
		java_lang_Thread_NORM_PRIORITY,
		threadStackSize,
		thread_malloc,
		thread_free,
		broadcastDeath,
		throwDeath);	
	assert(mainthread);
	gc_add_ref(mainthread);
}

static
void
TcreateFirst(Hjava_lang_Thread* tid)
{
	jthread_atexit(runfinalizer);

	/* set Java thread associated with main thread */
	GET_JTHREAD()->jlThread = tid;
	SET_COOKIE(tid);

	GC_WRITE(tid, mainthread);
	unhand(tid)->PrivateInfo = mainthread;

	unhand(tid)->stackOverflowError = 
		(Hjava_lang_Throwable*)StackOverflowError;
	unhand(tid)->needOnStack = STACK_HIGH;
}

static
void
Tcreate(Hjava_lang_Thread* tid, void* func)
{
	struct Hkaffe_util_Ptr* nativethread;
	int scheduler = 0;
#ifdef  CPU_INHERIT
	struct Hsys_Scheduler* psched = unhand(tid)->scheduler;

	if (psched)
		scheduler = (int) unhand(psched)->PrivateInfo;
#endif
	nativethread = (struct Hkaffe_util_Ptr*)jthread_create(
		unhand(tid)->priority,
		func,
		unhand(tid)->daemon,
		tid,
		scheduler,
		threadStackSize);
	GC_WRITE(tid, nativethread);
	unhand(tid)->PrivateInfo = nativethread;

	/* preallocate a stack overflow error for this thread in case it 
	   runs out */
	unhand(tid)->stackOverflowError = 
		(Hjava_lang_Throwable*)StackOverflowError;
	unhand(tid)->needOnStack = STACK_HIGH;
}

static  
void    
Tprio(Hjava_lang_Thread* tid, jint prio)
{       
	/* no native thread yet */
	if (unhand(tid)->PrivateInfo == 0)
		return;

	jthread_setpriority((jthread_t)unhand(tid)->PrivateInfo, prio);
}               

static          
void            
Tstop(Hjava_lang_Thread* tid)
{       
	jthread_stop((jthread_t)unhand(tid)->PrivateInfo);
}       

static          
void            
Tinterrupt(Hjava_lang_Thread* tid)
{       
#if 0
	jthread_interrupt((jthread_t)unhand(tid)->PrivateInfo);
#endif
}       

static  
bool    
Talive(Hjava_lang_Thread* tid)
{
	return jthread_alive((jthread_t)unhand(tid)->PrivateInfo);
}
        
static
jint    
Tframes(Hjava_lang_Thread* tid)
{
	return jthread_frames((jthread_t)unhand(tid)->PrivateInfo);
}       
 
static
void    
Tfinalize(Hjava_lang_Thread* tid)
{
	jthread_t jtid = (jthread_t)unhand(tid)->PrivateInfo;
	
	if (jtid != 0) {
		unhand(tid)->PrivateInfo = 0;
		jthread_destroy(jtid);
	}
}       

static          
Hjava_lang_Thread*      
TcurrentJava(void)      
{               
	return GET_COOKIE();
}

static          
void
TwalkThreads(void)
{               
	/* this is from gc-incremental.c */
	extern void walkMemory(void*);
	jthread_walkLiveThreads(walkMemory);
}

/*      
 * Walk the thread's internal context.
 */
static
void
TwalkThread(Hjava_lang_Thread* tid)
{
	void *from;
	unsigned len;
	jthread_t jtid = (jthread_t)unhand(tid)->PrivateInfo;

        if (jtid == 0)
                return;
 
        markObject(unhand(tid)->exceptObj);

	/* ask threading system what the interesting stack range is */
	jthread_extract_stack(jtid, &from, &len);

	/* and walk it */
	walkConservative(from, len);
}

/* called when unrolling exceptions */
static  
void*   
TnextFrame(void* fm)
{
#if defined(TRANSLATOR)
        exceptionFrame* nfm;

        nfm = (exceptionFrame*)(((exceptionFrame*)fm)->retbp);
	/* Note: this should obsolete the FRAMEOKAY macro */
        if (nfm && jthread_on_current_stack((void *)nfm->retbp)) {
                return (nfm);
        }
        else {
                return (0);
        }
#else
        vmException* nfm; 
        nfm = ((vmException*)fm)->prev;
        if (nfm != 0 && nfm->meth != (Method*)1) {
                return (nfm);
        }
        else {
                return (0);
        }
#endif 
}

static
Hjava_lang_Throwable*
TcheckStack(int left)
{
	int rc = jthread_stackcheck(left);

DBG(VMTHREAD,
	if (rc == false) {
		dprintf("%s doesn't have %d bytes left\n", 
			nameThread(GET_COOKIE()), left);
	}
   )
	return (Hjava_lang_Throwable*)
		(rc == true ? 0 : unhand(TcurrentJava())->stackOverflowError);
}

/*
 * implementation of the locking subsystem based on jlocks
 *
 * Note that we keep track of lk->holder, and its type is void*.
 */
static
void  
Linit(iLock* lk)
{
	static bool first = true;
	static jmutex first_mutex;
	static jcondvar first_condvar;

	/* The first lock init is for the memory manager - so we can't
	 * use it yet.  Allocate from static space.
	 */
	if (first == true) {
		first = false;
		lk->mux = &first_mutex;
		lk->cv = &first_condvar;
	}
	else {
		lk->mux = thread_malloc(sizeof(jmutex));
		lk->cv = thread_malloc(sizeof(jcondvar));
	}
	jmutex_initialise(lk->mux);
	jcondvar_initialise(lk->cv);
}               
                
static  
void    
Llock(iLock* lk)
{
	jmutex_lock(lk->mux);
	lk->holder = (void *)jthread_current();
}

static  
void    
Lunlock(iLock* lk)
{       
	lk->holder = 0;
	jmutex_unlock(lk->mux);
}

static
void
Lwait(iLock* lk, jlong timeout)
{
	int count;
	/*
	 * We must reacquire the Java lock before we're ready to die
	 */
	jthread_disable_stop();
	count = lk->count;
	lk->count = 0;
	jcondvar_wait(lk->cv, lk->mux, timeout);
	lk->holder = (void *)jthread_current();
	lk->count = count;

	/* now it's safe to start dying */ 
	jthread_enable_stop();
}
 
static  
void    
Lsignal(iLock* lk)
{
	jcondvar_signal(lk->cv, lk->mux);
}       
        
static          
void
Lbroadcast(iLock* lk)
{
	jcondvar_broadcast(lk->cv, lk->mux);
}

void    
Tspinon(void* arg)
{       
	intsDisable();
}       
        
void            
Tspinoff(void* arg)
{               
	intsRestore();
}               

void    
Tspinoffall(void* arg)
{       
	intsRestoreAll();
}       
        
/*
 * Wait for a child process.
 */
static int
my_waitpid(int wpid, int* status, int options)
{
	SignalError("java.lang.InternalError", "no waitpid() available");

	return -1;
}

/*
 * socket connect.
 */
static int
my_connect(int fd, struct sockaddr* addr, size_t len)
{
        int r = connect(fd, addr, len);
        /* annul EALREADY error */
        if (r == -1 && errno == EALREADY)
                r = 0;
        return r;
}

static int 
my_mkdir(const char *path, int m)
{
	return mkdir(path, m);
}

static int
my_forkexec(char *argv[], char *env[], int fd[4])
{
	unimp("fork yourself.");
}

/*
 * Define the thread interface.
 */
ThreadInterface Kaffe_ThreadInterface = {

        Tinit,
        TcreateFirst,
        Tcreate,
        jthread_sleep,
        jthread_yield,
        Tprio, 
        Tstop,
	Tinterrupt,
        jthread_exit,
        Talive,
        Tframes,
        Tfinalize,
        TcurrentJava,
	(void *(*)())jthread_current,
        TwalkThreads,
        TwalkThread,
        TnextFrame,
	TcheckStack,
};

/*
 * Define the lock interface.
 */
LockInterface Kaffe_LockInterface = {

        Linit,
        Llock,
        Lunlock,
        Lwait,
        Lsignal,
        Lbroadcast,

        Tspinon,
        Tspinoff,

};

void
nuts(void)
{
	panic("nuts");
}

SystemCallInterface Kaffe_SystemCallInterface = {

        (int (*)(const char *, int, int))open,	
        read,	
        write, 
        lseek,
        close,
        fstat,
        stat,

        my_mkdir,		/* the real mkdir takes a mode_t */
        rmdir,
        rename,
        remove,

        socket,
        my_connect,
        (int (*)(int, struct sockaddr *, size_t*))accept, 
        read,	
        recvfrom,
        write, 
        sendto,	
        setsockopt,	
        getsockopt,
        getsockname, 
        getpeername,

        select,	

	my_forkexec,
        my_waitpid,
	nuts,
};

#ifdef CPU_INHERIT
void
TcreateFixedPrioScheduler(struct Hsys_Scheduler* psched)
{
	struct Hkaffe_util_Ptr* nativethread;
	int			prio    = (int) unhand(psched)->priority;
	int			preempt = (int) unhand(psched)->preemptable;
	struct Hsys_Scheduler*  parent  = unhand(psched)->parent;
	int			ptid    = 0;

	if (parent)
		ptid = (int) unhand(parent)->PrivateInfo;

	nativethread = (struct Hkaffe_util_Ptr *)
		jthread_createFixedPrioScheduler(ptid, prio, preempt);

	unhand(psched)->PrivateInfo = nativethread;
}

TcreateRateMonoScheduler(struct Hsys_Scheduler* psched)
{
	struct Hkaffe_util_Ptr* nativethread;
	int			prio    = (int) unhand(psched)->priority;
	int			preempt = (int) unhand(psched)->preemptable;
	struct Hsys_Scheduler*  parent  = unhand(psched)->parent;
	int			ptid    = 0;

	if (parent)
		ptid = (int) unhand(parent)->PrivateInfo;

	nativethread = (struct Hkaffe_util_Ptr *)
		jthread_createRateMonoScheduler(ptid, prio, preempt);

	unhand(psched)->PrivateInfo = nativethread;
}

TcreateLotteryScheduler(struct Hsys_Scheduler* psched)
{
	struct Hkaffe_util_Ptr* nativethread;
	int			prio    = (int) unhand(psched)->priority;
	int			preempt = (int) unhand(psched)->preemptable;
	struct Hsys_Scheduler*  parent  = unhand(psched)->parent;
	int			ptid    = 0;

	if (parent)
		ptid = (int) unhand(parent)->PrivateInfo;

	nativethread = (struct Hkaffe_util_Ptr *)
		jthread_createLotteryScheduler(ptid, prio, preempt);

	unhand(psched)->PrivateInfo = nativethread;
}

TcreateStrideScheduler(struct Hsys_Scheduler* psched)
{
	struct Hkaffe_util_Ptr* nativethread;
	int			prio    = (int) unhand(psched)->priority;
	int			preempt = (int) unhand(psched)->preemptable;
	struct Hsys_Scheduler*  parent  = unhand(psched)->parent;
	int			ptid    = 0;

	if (parent)
		ptid = (int) unhand(parent)->PrivateInfo;

	nativethread = (struct Hkaffe_util_Ptr *)
		jthread_createStrideScheduler(ptid, prio, preempt);

	unhand(psched)->PrivateInfo = nativethread;
}

jint
Tcputime(struct Hjava_lang_Thread* tid)
{
	return jthread_cputime((jthread_t)unhand(tid)->PrivateInfo);	
}
#endif
